/*
 * manage HTTP access
 *
 * Copyright(c) 2008 olyutorskii
 * $Id: ServerAccess.java 337 2009-01-15 09:46:55Z olyutorskii $
 */

package jp.sourceforge.jindolf;

import java.awt.Image;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.lang.ref.SoftReference;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.text.NumberFormat;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import javax.imageio.ImageIO;

/**
 * 国ごとの人狼BBSサーバとの通信を一手に引き受ける。
 */
public class ServerAccess{

    private static final String USER_AGENT = HttpUtils.getUserAgentName();
    private static final String ENCODING_HTML = "Shift_JIS";
    // TODO MS932とかWindows31Jとかのほうがいいか？
    private static final Charset CHARSET_SJIS = Charset.forName(ENCODING_HTML);
    private static final int BUF_SIZE = 1024;
    private static final String JINRO_CGI = "./index.rb";
    private static final Map<String, SoftReference<Image>> imageCache;
    private static final NumberFormat throughputFormat;
    private static final NumberFormat sizeFormat;

    static{
        Map<String, SoftReference<Image>> cache
                = new HashMap<String, SoftReference<Image>>();
        imageCache = Collections.synchronizedMap(cache);
        throughputFormat = NumberFormat.getInstance();
        throughputFormat.setMaximumFractionDigits(1);
        throughputFormat.setMinimumFractionDigits(1);
        throughputFormat.setGroupingUsed(true);
        sizeFormat = NumberFormat.getInstance();
        sizeFormat.setGroupingUsed(true);
    }

    /**
     * 画像キャッシュを検索する。
     * @param key キー
     * @return キャッシュされた画像。キャッシュされていなければnull。
     */
    private static Image getImageCache(String key){
        if(key == null) return null;

        Image image;

        synchronized(imageCache){
            SoftReference<Image> ref = imageCache.get(key);
            if(ref == null) return null;

            Object referent = ref.get();
            if(referent == null){
                imageCache.remove(key);
                return null;
            }

            image = (Image) referent;
        }

        return image;
    }

    /**
     * 画像キャッシュに登録する。
     * @param key キー
     * @param image キャッシュしたい画像。
     */
    private static void putImageCache(String key, Image image){
        if(key == null || image == null) return;

        synchronized(imageCache){
            if(getImageCache(key) != null) return;
            SoftReference<Image> ref = new SoftReference<Image>(image);
            imageCache.put(key, ref);
        }

        return;
    }

    /**
     * 与えられた文字列に対し「application/x-www-form-urlencoded」符号化を行う。
     * この符号化はHTTPのPOSTメソッドで必要になる。
     * この処理は、一般的なPC用Webブラウザにおける、
     * Shift_JISで書かれたHTML文書のFORMタグに伴う
     * submit処理を模倣する。
     * @param formData 元の文字列
     * @return 符号化された文字列
     */
    public static String formEncode(String formData){
        if(formData == null){
            return null;
        }
        String result;
        try{
            result = URLEncoder.encode(formData, ENCODING_HTML);
        }catch(UnsupportedEncodingException e){
            assert false;
            result = null;
        }
        return result;
    }

    /**
     * 配列版formEncode
     * @param formData 元の文字列
     * @return 符号化された文字列
     */
    public static String formEncode(char[] formData){
        return formEncode(new String(formData));
    }

    /**
     * 「Shift_JIS」でエンコーディングされた入力ストリームから文字列を生成する。
     * @param is 入力ストリーム
     * @return 文字列
     * @throws java.io.IOException 入出力エラー（おそらくネットワーク関連）
     */
    public static CharSequence downloadHTMLStream(InputStream is)
            throws IOException{
        StringBuilder result = new StringBuilder();

        Reader reader;
        // TODO 「～」が化ける。他にもいろいろあるはず。
        reader = new InputStreamReader(is, CHARSET_SJIS);
        reader = new BufferedReader(reader, BUF_SIZE);

        char[] buffer = new char[BUF_SIZE];
        for(;;){
            int chs = reader.read(buffer);
            if(chs <= 0){
                break;
            }
            result.append(buffer, 0, chs);
        }
        reader.close();

        return result;
    }

    /**
     * ネットワークのスループット報告用文字列を生成する。
     * @param size 転送サイズ(バイト数)
     * @param nano 所要時間(ナノ秒)
     * @return スループット文字列
     */
    private static String throughput(long size, long nano){
        double sec = ((double)nano) / (1000.0*1000.0*1000.0);
        double rate = ((double)size) / sec;

        String unit = "";
        if(rate >= 1500.0){
            rate /= 1000.0;
            unit = "K";
        }
        if(rate >= 1500.0){
            rate /= 1000.0;
            unit = "M";
        }

        String result =  sizeFormat.format(size) + "Bytes "
                       + throughputFormat.format(rate) + unit
                       + "Bytes/sec";
        return result;
    }

    private final URL baseURL;
    private final URL cgiURL;
    private long lastServerMs;
    private long lastLocalMs;
    private long lastSystemMs;
    private AccountCookie cookieAuth = null;
    private String encodedUserID = null;

    /**
     * 人狼BBSサーバとの接続管理を生成する。
     * この時点ではまだ通信は行われない。
     * @param baseURL 国別のベースURL
     */
    public ServerAccess(URL baseURL){
        this.baseURL = baseURL;
        try{
            this.cgiURL = new URL(this.baseURL, JINRO_CGI);
        }catch(MalformedURLException e){
            assert false;
            throw new Error();
        }
        return;
    }

    /**
     * 国のベースURLを返す。
     * @return ベースURL
     */
    public URL getBaseURL(){
        return this.baseURL;
    }

    /**
     * 与えられたクエリーとCGIのURLから新たにURLを合成する。
     * @param query クエリー
     * @return 新たなURL
     */
    protected URL getQueryURL(String query){
        if(query.length() >= 1 && query.charAt(0) != '?'){
            return null;
        }

        URL result;
        try{
            result = new URL(this.cgiURL, query);
        }catch(MalformedURLException e){
            assert false;
            return null;
        }
        return result;
    }

    /**
     * 与えられたクエリーを用いてHTMLデータを取得する。
     * @param query HTTP-GET クエリー
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    protected CharSequence downloadHTML(String query)
            throws IOException{
        URL url = getQueryURL(query);
        HttpURLConnection connection
                = (HttpURLConnection) url.openConnection();
        connection.setRequestProperty("Accept", "*/*");
        connection.setRequestProperty("User-Agent", USER_AGENT);
        connection.setUseCaches(false);
        connection.setInstanceFollowRedirects(false);
        connection.setDoInput(true);
        connection.setRequestMethod("GET");

        AccountCookie cookie = this.cookieAuth;
        if(cookie != null){
            if(shouldAccept(url, cookie)){
                connection.setRequestProperty(
                        "Cookie",
                        "login=" + cookie.getLoginData());
            }else{
                clearAuthentication();
            }
        }

        long nanoLap = System.nanoTime();

        connection.connect();

        updateLastAccess(connection);

        int responseCode       = connection.getResponseCode();
        String responseMessage = connection.getResponseMessage();
        String logMessage =   "GET " + url
                            + " [" + responseCode
                            + " " + responseMessage + "]";

        if(responseCode != HttpURLConnection.HTTP_OK){
            logMessage =  "発言のダウンロードに失敗しました。"
                        + " " + logMessage;
            Jindolf.logger.log(Level.WARNING, logMessage);
            return null;
        }

        String charset = HttpUtils.getHTMLCharset(connection);
        if(!charset.equalsIgnoreCase(ENCODING_HTML)){
            return null;
        }

        InputStream stream = connection.getInputStream();
        stream = new BufferedInputStream(stream, 512);
        CountInputStream cistream = new CountInputStream(stream);
        CharSequence html = downloadHTMLStream(cistream);
        cistream.close();

        connection.disconnect();

        nanoLap = System.nanoTime() - nanoLap;
        long htmlSize = cistream.getCount();
        String throughput = throughput(htmlSize, nanoLap);
        logMessage = logMessage + " " + throughput;
        Jindolf.logger.info(logMessage);

        return html;
    }

    /**
     * 絶対または相対URLの指すパーマネントなイメージ画像をダウンロードする。
     * @param url 画像URL文字列
     * @return 画像イメージ
     * @throws java.io.IOException ネットワークエラー
     */
    public Image downloadImage(String url) throws IOException{
        URL absolute;
        try{
            URL base = getBaseURL();
            absolute = new URL(base, url);
        }catch(MalformedURLException e){
            assert false;
            return null;
        }

        Image image;
        image = getImageCache(absolute.toString());
        if(image != null) return image;

        HttpURLConnection connection
                = (HttpURLConnection) absolute.openConnection();
        connection.setRequestProperty("Accept", "*/*");
        connection.setRequestProperty("User-Agent", USER_AGENT);
        connection.setUseCaches(true);
        connection.setInstanceFollowRedirects(true);
        connection.setDoInput(true);
        connection.setRequestMethod("GET");

        long nanoLap = System.nanoTime();

        connection.connect();

        int responseCode       = connection.getResponseCode();
        String responseMessage = connection.getResponseMessage();
        String logMessage =   "GET " + absolute
                            + " [" + responseCode
                            + " " + responseMessage + "]";

        if(responseCode != HttpURLConnection.HTTP_OK){
            logMessage =  "イメージのダウンロードに失敗しました。"
                        + " " + logMessage;
            Jindolf.logger.log(Level.WARNING, logMessage);
            return null;
        }

        InputStream stream = connection.getInputStream();
        stream = new BufferedInputStream(stream, 512);
        CountInputStream cistream = new CountInputStream(stream);
        image = ImageIO.read(cistream);
        cistream.close();

        connection.disconnect();

        nanoLap = System.nanoTime() - nanoLap;
        long imageSize = cistream.getCount();
        String throughput = throughput(imageSize, nanoLap);
        logMessage = logMessage + " " + throughput;
        Jindolf.logger.info(logMessage);

        putImageCache(absolute.toString(), image);

        return image;
    }

    /**
     * 指定された認証情報をPOSTする。
     * @param authData 認証情報
     * @return 認証情報が受け入れられたらtrue
     * @throws java.io.IOException ネットワークエラー
     */
    protected boolean postAuthData(String authData) throws IOException{
        URL url = getBaseURL();
        HttpURLConnection connection
                = (HttpURLConnection) url.openConnection();
        connection.setRequestProperty("Accept", "*/*");
        connection.setRequestProperty("User-Agent", USER_AGENT);
        connection.setUseCaches(false);
        connection.setInstanceFollowRedirects(false);
        connection.setDoInput(true);
        connection.setDoOutput(true);
        connection.setRequestMethod("POST");

        byte[] authBytes = authData.getBytes();

        OutputStream os = connection.getOutputStream();

        long nanoLap = System.nanoTime();

        os.write(authBytes);
        os.flush();
        os.close();

        updateLastAccess(connection);

        int responseCode       = connection.getResponseCode();
        String responseMessage = connection.getResponseMessage();
        String logMessage =   "POST " + url
                            + " [" + responseCode
                            + " " + responseMessage + "]";

        if(responseCode != HttpURLConnection.HTTP_MOVED_TEMP){    // 302
            logMessage =  "認証情報の送信に失敗しました。"
                        + " " + logMessage;
            Jindolf.logger.log(Level.WARNING, logMessage);
            return false;
        }

        connection.disconnect();

        nanoLap = System.nanoTime() - nanoLap;
        long sendSize = authBytes.length;
        String throughput = throughput(sendSize, nanoLap);
        logMessage = logMessage + " " + throughput;
        Jindolf.logger.info(logMessage);

        AccountCookie loginCookie = AccountCookie.createCookie(connection);
        if(loginCookie == null){
            return false;
        }

        setAuthentication(loginCookie);

        return true;
    }

    /**
     * トップページのHTMLデータを取得する。
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    public CharSequence getHTMLTopPage() throws IOException{
        return downloadHTML("");
    }

    /**
     * 国に含まれる村一覧HTMLデータを取得する。
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    public CharSequence getHTMLLandList() throws IOException{
        return downloadHTML("?cmd=log");
    }

    /**
     * 指定された村のPeriod一覧のHTMLデータを取得する。
     * 現在ゲーム進行中の村にも可能。
     * ※ 古国では使えないよ！
     * @param village 村
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    public CharSequence getHTMLBoneHead(Village village) throws IOException{
        String villageID = village.getVillageID();
        return downloadHTML("?vid=" + villageID + "&meslog=");
    }

    /**
     * 指定された村の最新PeriodのHTMLデータをロードする。
     * 既にGAMEOVERの村ではPeriod一覧のHTMLデータとなる。
     * @param village 村
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    public CharSequence getHTMLVillage(Village village) throws IOException{
        String villageID = village.getVillageID();
        return downloadHTML("?vid=" + villageID);
    }

    /**
     * 指定されたPeriodのHTMLデータをロードする
     * @param period Period
     * @return HTMLデータ
     * @throws java.io.IOException ネットワークエラー
     */
    public CharSequence getHTMLPeriod(Period period) throws IOException{
        String query = period.getCGIQuery();
        return downloadHTML(query);
    }

    /**
     * 最終アクセス時刻を更新する。
     * @param connection HTTP接続
     */
    public void updateLastAccess(HttpURLConnection connection){
        this.lastServerMs = connection.getDate();
        this.lastLocalMs = System.currentTimeMillis();
        this.lastSystemMs = System.nanoTime() / (1000 * 1000);
    }

    /**
     * 指定したURLに対しCookieを送っても良いか否か判定する。
     * 判別材料は Cookie の寿命とパス指定のみ。
     * @param url URL
     * @param cookie Cookie
     * @return 送ってもよければtrue
     */
    private static boolean shouldAccept(URL url, AccountCookie cookie){
        if(cookie.hasExpired()){
            return false;
        }

        String urlPath = url.getPath();
        String cookiePath = cookie.getPathURI().getPath();

        if( ! urlPath.startsWith(cookiePath) ){
            return false;
        }

        return true;
    }

    /**
     * 現在ログイン中か否か判別する。
     * @return ログイン中ならtrue
     */
    // TODO interval call
    public boolean hasLoggedIn(){
        AccountCookie cookie = this.cookieAuth;
        if(cookie == null){
            return false;
        }
        if(cookie.hasExpired()){
            clearAuthentication();
            return false;
        }
        return true;
    }

    /**
     * 与えられたユーザIDとパスワードでログイン処理を行う。
     * @param userID ユーザID
     * @param password パスワード
     * @return ログインに成功すればtrue
     * @throws java.io.IOException ネットワークエラー
     */
    public final boolean login(String userID, char[] password)
            throws IOException{
        if(hasLoggedIn()){
            return true;
        }

        String id = formEncode(userID);
        if(id == null || id.length() <= 0){
            return false;
        }

        String pw = formEncode(password);
        if(pw == null || pw.length() <= 0){
            return false;
        }

        this.encodedUserID = id;

        String redirect = formEncode("&#bottom");   // TODO ほんとに必要？

        StringBuilder postData = new StringBuilder();
        postData.append("cmd=login");
        postData.append('&').append("cgi_param=").append(redirect);
        postData.append('&').append("user_id=").append(id);
        postData.append('&').append("password=").append(pw);

        boolean result;
        try{
            result = postAuthData(postData.toString());
        }catch(IOException e){
            clearAuthentication();
            throw e;
        }

        return result;
    }

    /**
     * ログアウト処理を行う。
     * @throws java.io.IOException
     */
    // TODO シャットダウンフックでログアウトさせようかな…
    public void logout() throws IOException{
        if(!hasLoggedIn()){
            return;
        }
        if(this.encodedUserID == null){
            clearAuthentication();
            return;
        }

        String redirect = formEncode("&#bottom"); // TODO 必要？

        StringBuilder postData = new StringBuilder();
        postData.append("cmd=logout");
        postData.append('&').append("cgi_param=").append(redirect);
        postData.append('&').append("user_id=").append(this.encodedUserID);

        try{
            postAuthData(postData.toString());
        }finally{
            clearAuthentication();
        }

        return;
    }

    /**
     * 認証情報クリア
     */
    // TODO タイマーでExpire date の時刻にクリアしたい。
    protected void clearAuthentication(){
        this.cookieAuth = null;
        this.encodedUserID = null;
        return;
    }

    /**
     * 認証情報のセット
     * @param cookie 認証Cookie
     */
    private void setAuthentication(AccountCookie cookie){
        this.cookieAuth = cookie;
        return;
    }

    // TODO JRE1.6対応するときに HttpCookie, CookieManager 利用へ移行したい。

}
